/*==============================================================================


   Purpose:    small test program for hpm


   Author:     Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   02.07.1999


==============================================================================*/

/* This is a test program for hpm. Execute YOUR_BINARY_PATH/hpm a.out after
   compiling this program.
 */


#include <stdio.h>

#include "processor_number.h"

/* dimension of test arrays */
#define N 1000

#define ITER_COUNT 100

/*                                                                            */
/*============================================================================*/
/* matrix addition as a simple test program.                                  */


static void
matadd(double a[N][N], double b[N][N], double c[N][N], int n)

{
  int i, j;


  for (i = 0; i < n; ++i)
    for (j = 0; j < n; ++j)
      a[i][j] = b[i][j] + c[i][j];
}


int main(int argc, char **argv)
{
  double a[N][N], b[N][N], c[N][N];
  int i,j,n=N;

  printf("starting %s on processor %d\n", argv[0], PCL_processor_number());

  if(argc > 1)
    {
      if((sscanf(argv[1], "%d", &n) != 1)
	 || (n > N))
	printf("wrong argument (%s)\n", argv[1]);
    }

  /* initialize test arrays */
  for (i = 0; i < n; ++i)
    for (j = 0; j < n; ++j)
      a[i][j] = b[i][j] = c[i][j] = 1.0;

  /* do some initial work */
  for(i = 0; i < ITER_COUNT; ++i)
    matadd(a, b, c, n);
  
  printf("stopping %s (%d FP operations)\n", argv[0], n*n*ITER_COUNT);

  return 0;
}
