/*
 * x86 machine specific register device 
 * originally written by
 * Stephan Meyer <Stephan.Meyer@pobox.com>
 * and still somewhat supported.
 * When I find the time, I plan to implement a machine-independent
 * sort-of-general interface to stuff like that.
 *
 * Changes and documentation Chris Wedgwood <chris@cybernet.co.nz>
 *
 * updates for 2.1.62/3 by John Gotts <jgotts@umich.edu>
 *
 * This source code is covered by the GNU GPL 2.0.
 *
 * ***********
 * * NOTE!!! *
 * ***********
 *
 * The hi & lo parts of the msr register have been swapped from what
 * was previously pre-released. This is the 'right way' and allows
 * you to read into the address of a long long.
 *
 * This will break ALL old code. This was not done so much as to make
 * it the right way before it's too late, but more to break
 * everyone's code and piss people off.
 *
 * Anyhow, the SMP processor flag is of limited use. I guess we could
 * possibly use something like smp_message_pass and dick about, but
 * it's almost certainly more trouble than its worth.
 *
 * If you really want good event info for a variety of processors on
 * both uniprocessor and smp machines, then you need to hack the
 * scheduler and keep task by task counters.
 */

/* need to do "mknod /dev/msr c 10 142" to make the device */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/smp.h>
#include <asm/processor.h>

#define MSR_MINOR 142

/*
 * We only allow one instance of the device open at a time, this
 * should avoid funny things happening with certain msr's...
 *
 * some may argue that this may not the the 'right' behaviour. Life
 * is short.
 */

static int is_open = 0;

static int msr_open(struct inode *finode, struct file *file)
{
	if(is_open)
		return -EBUSY;
#ifdef	MODULE
	MOD_INC_USE_COUNT;
#endif
	is_open = !0;

	return	0;
}

static int msr_close(struct file *file)
{
#ifdef	MODULE
	MOD_DEC_USE_COUNT;
#endif
	return is_open = 0;
}

/*
 * - reading less than 8 bytes will read nothing.
 * - reading 8-11 bytes will read a single msr
 * - reading 12 bytes will read thr msr and place the processor id 
 *   last dword.
 * - reading more than 12 bytes is the same as reading 12 bytes.
 *
 * All this checking and stuff may affect certain msr's, but it
 * should be below the noise introduced by the syscall overhead.
 */
 
static ssize_t msr_read(struct file *file, char *buf, size_t count,
				loff_t *off)
{
	unsigned long int p=file->f_pos,hi,lo,pn;

	if(count<8)
		return	0;
		
	__asm__ __volatile__ (
		"rdmsr"
		:"=a" (lo), "=d" (hi)
		:"c" (p)
		:"eax","ecx","edx");

	if(copy_to_user(buf,&lo,4))
		return -EFAULT;
	if(copy_to_user(buf+4,&hi,4))
		return -EFAULT;

	if(count<12)
		return 8;
	
	pn = smp_processor_id();

	if(copy_to_user(buf+8,&pn,4))
		return -EFAULT;
	
	return	12;	
}

/* writing less than 8 bytes will write nothing -cw */

static ssize_t msr_write(struct file *file, const char *buf, size_t count,
				loff_t *ppos)
{
	unsigned long p=file->f_pos,hi,lo;

	if(count<8)
		return	0;
		
	if(copy_from_user(&lo,buf,4))
		return -EFAULT;
	if(copy_from_user(&hi,buf+4,4))
		return -EFAULT;

	__asm__ __volatile__ (
		"wrmsr"
		:
		:"c" (p), "a" (lo), "d" (hi)
		:"eax","ecx","edx");

	return 8;
}

static loff_t msr_seek(struct file *file, loff_t offset, int origin)
{
	switch (origin) {
		case 0:	 /* SEEK_SET */
			file->f_pos=offset;
			return file->f_pos;
		case 1:	 /* SEEK_CUR */
			file->f_pos+=offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static struct file_operations msr_fops = {
	msr_seek,
	msr_read,
	msr_write,
	NULL,
	NULL,
	NULL,
	NULL,
	msr_open,
	msr_close,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

static struct miscdevice msr_device = {
	MSR_MINOR,
	"msr",
	&msr_fops
};


#ifdef MODULE

MODULE_AUTHOR("Stephan Meyer <Stephan.Meyer@pobox.com>");
MODULE_DESCRIPTION("Allows access to the x86 msr registers");

int init_module(void)
{
        unsigned int signature, feature_flags, family;

        asm("cpuid\n\t"
	    : "=a" (signature), "=d"(feature_flags)
	    : "a" (1)
	    : "%eax", "%ebx", "%ecx", "%edx");

	family = (signature >> 8) & 0xf;
	if ((family >= 6)
	    || ((family == 5) && (((feature_flags >> 4) & 0x1) == 1)))
	  {
	    /* either PentiumPro or better (family>=6) or Pentium MMX */
	    printk(KERN_INFO "\nEnabling user-level RDPMC.\n");
	    /* Enable RDPMC at any protection level */
	    asm("movl %cr4,%eax\n"); 
	    asm("orl $256,%eax\n");
	    asm("movl %eax,%cr4\n");
	  }

	misc_register(&msr_device);
	printk(KERN_DEBUG "MSR: device installed\n");
	
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&msr_device);
	printk(KERN_DEBUG "MSR: device uninstalled\n");
}

#else
	
void msr_init (void)
{
        unsigned int signature, feature_flags, family;

	if ((!have_cpuid) || (!(x86_capability & 32))) {
		printk(KERN_ERR "MSR: msr not supported!\n");
		return;
	}

        asm("cpuid\n\t"
	    : "=a" (signature), "=d"(feature_flags)
	    : "a" (1)
	    : "%eax", "%ebx", "%ecx", "%edx");

	family = (signature >> 8) & 0xf;
	if ((family >= 6)
	    || ((family == 5) && (((feature_flags >> 4) & 0x1) == 1)))
	  {
	    /* either PentiumPro or better (family>=6) or Pentium MMX */
	    printk(KERN_INFO "\nEnabling user-level RDPMC.\n");
	    /* Enable RDPMC at any protection level */
	    asm("movl %cr4,%eax\n"); 
	    asm("orl $256,%eax\n");
	    asm("movl %eax,%cr4\n");
	  }

	if ((current_cpu_data.cpuid_level==-1) || 
	    (!(current_cpu_data.x86_capability & X86_FEATURE_MSR))) {
		printk("MSR: msr not supported!\n");
		return;
	}

	printk(KERN_DEBUG "MSR: device installed.\n");
	misc_register(&msr_device);
}

#endif
